#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import sys
from unittest import mock

from keystoneauth1 import fixture


AUTH_TOKEN = "foobar"
AUTH_URL = "http://0.0.0.0"
USERNAME = "itchy"
USER_ID = "2354b7c1-f681-4c39-8003-4fe9d1eabb65"
PASSWORD = "scratchy"
PROJECT_NAME = "poochie"
PROJECT_ID = "30c3da29-61f5-4b7b-8eb2-3d18287428c7"
REGION_NAME = "richie"
INTERFACE = "catchy"
VERSION = "3"

SERVICE_PROVIDER_ID = "bob"

TEST_RESPONSE_DICT = fixture.V2Token(token_id=AUTH_TOKEN, user_name=USERNAME)
_s = TEST_RESPONSE_DICT.add_service('identity', name='keystone')
_s.add_endpoint(AUTH_URL + ':5000/v2.0')
_s = TEST_RESPONSE_DICT.add_service('network', name='neutron')
_s.add_endpoint(AUTH_URL + ':9696')
_s = TEST_RESPONSE_DICT.add_service('compute', name='nova')
_s.add_endpoint(AUTH_URL + ':8774/v2')
_s = TEST_RESPONSE_DICT.add_service('image', name='glance')
_s.add_endpoint(AUTH_URL + ':9292')
_s = TEST_RESPONSE_DICT.add_service('object', name='swift')
_s.add_endpoint(AUTH_URL + ':8080/v1')

TEST_RESPONSE_DICT_V3 = fixture.V3Token(user_name=USERNAME)
TEST_RESPONSE_DICT_V3.set_project_scope()

TEST_VERSIONS = fixture.DiscoveryList(href=AUTH_URL)


def to_unicode_dict(catalog_dict):
    """Converts dict to unicode dict"""
    if isinstance(catalog_dict, dict):
        return {
            to_unicode_dict(key): to_unicode_dict(value)
            for key, value in catalog_dict.items()
        }
    elif isinstance(catalog_dict, list):
        return [to_unicode_dict(element) for element in catalog_dict]
    elif isinstance(catalog_dict, str):
        return catalog_dict + ""
    else:
        return catalog_dict


class FakeStdout:
    def __init__(self):
        self.content = []

    def write(self, text):
        self.content.append(text)

    def make_string(self):
        result = ''
        for line in self.content:
            result = result + line
        return result


class FakeLog:
    def __init__(self):
        self.messages = {}

    def debug(self, msg):
        self.messages['debug'] = msg

    def info(self, msg):
        self.messages['info'] = msg

    def warning(self, msg):
        self.messages['warning'] = msg

    def error(self, msg):
        self.messages['error'] = msg

    def critical(self, msg):
        self.messages['critical'] = msg


class FakeApp:
    def __init__(self, _stdout, _log):
        self.stdout = _stdout
        self.client_manager = None
        self.api_version = {}
        self.stdin = sys.stdin
        self.stdout = _stdout or sys.stdout
        self.stderr = sys.stderr
        self.log = _log


class FakeOptions:
    def __init__(self, **kwargs):
        self.os_beta_command = False


class FakeClientManager:
    def __init__(self):
        self.compute = None
        self.identity = None
        self.image = None
        self.object_store = None
        self.volume = None
        self.network = None
        self.session = None
        self.auth_ref = None
        self.auth_plugin_name = None

    def get_configuration(self):
        return {
            'auth': {
                'username': USERNAME,
                'password': PASSWORD,
                'token': AUTH_TOKEN,
            },
            'region': REGION_NAME,
            'identity_api_version': VERSION,
        }


class FakeModule:
    def __init__(self, name, version):
        self.name = name
        self.__version__ = version
        # Workaround for openstacksdk case
        self.version = mock.Mock()
        self.version.__version__ = version


class FakeResource:
    def __init__(self, manager=None, info=None, loaded=False, methods=None):
        """Set attributes and methods for a resource.

        :param manager:
            The resource manager
        :param Dictionary info:
            A dictionary with all attributes
        :param bool loaded:
            True if the resource is loaded in memory
        :param Dictionary methods:
            A dictionary with all methods
        """
        info = info or {}
        methods = methods or {}

        self.__name__ = type(self).__name__
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._add_methods(methods)
        self._loaded = loaded

    def _add_details(self, info):
        for k, v in info.items():
            setattr(self, k, v)

    def _add_methods(self, methods):
        """Fake methods with MagicMock objects.

        For each <@key, @value> pairs in methods, add an callable MagicMock
        object named @key as an attribute, and set the mock's return_value to
        @value. When users access the attribute with (), @value will be
        returned, which looks like a function call.
        """
        for name, ret in methods.items():
            method = mock.MagicMock(return_value=ret)
            setattr(self, name, method)

    def __repr__(self):
        reprkeys = sorted(
            k for k in self.__dict__.keys() if k[0] != '_' and k != 'manager'
        )
        info = ", ".join(f"{k}={getattr(self, k)}" for k in reprkeys)
        return f"<{self.__class__.__name__} {info}>"

    def keys(self):
        return self._info.keys()



#   Copyright 2019 Red Hat, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import argparse
import functools
import sys
from unittest import mock

from osc_lib.tests import utils as test_utils
from osc_lib.utils import tags


def help_enhancer(_h):
    """A simple helper to validate the ``enhance_help`` kwarg."""
    return ''.join(reversed(_h))


class TestTags(test_utils.TestCase):
    def test_add_tag_filtering_option_to_parser(self):
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        tags.add_tag_filtering_option_to_parser(parser, 'test')

        parsed_args = parser.parse_args(
            [
                '--tags',
                'tag1,tag2',
                '--any-tags',
                'tag4',
                '--not-tags',
                'tag5',
                '--not-any-tags',
                'tag6',
            ]
        )

        actual = getattr(parsed_args, 'tags', [])
        expected = ['tag1', 'tag2']
        self.assertCountEqual(expected, actual)

        actual = getattr(parsed_args, 'any_tags', [])
        expected = ['tag4']
        self.assertCountEqual(expected, actual)

        actual = getattr(parsed_args, 'not_tags', [])
        expected = ['tag5']
        self.assertCountEqual(expected, actual)

        actual = getattr(parsed_args, 'not_any_tags', [])
        expected = ['tag6']
        self.assertCountEqual(expected, actual)

    def test_get_tag_filtering_args(self):
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        tags.add_tag_filtering_option_to_parser(parser, 'test')

        parsed_args = parser.parse_args(
            [
                '--tags',
                'tag1,tag2',
                '--any-tags',
                'tag4',
                '--not-tags',
                'tag5',
                '--not-any-tags',
                'tag6',
            ]
        )
        expected = {
            'tags': 'tag1,tag2',
            'any_tags': 'tag4',
            'not_tags': 'tag5',
            'not_any_tags': 'tag6',
        }
        args = {}
        tags.get_tag_filtering_args(parsed_args, args)
        self.assertEqual(expected, args)

    def test_add_tag_option_to_parser_for_create(self):
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        tags.add_tag_option_to_parser_for_create(parser, 'test')

        # Test that --tag and --no-tag are mutually exclusive
        self.assertRaises(
            SystemExit, parser.parse_args, ['--tag', 'tag1', '--no-tag']
        )

        parsed_args = parser.parse_args(['--tag', 'tag1'])
        actual = getattr(parsed_args, 'tags', [])
        expected = ['tag1']
        self.assertCountEqual(expected, actual)

        parsed_args = parser.parse_args(['--no-tag'])
        actual = getattr(parsed_args, 'no-tag', [])
        expected = []
        self.assertCountEqual(expected, actual)

    def test_add_tag_option_to_parser_for_set(self):
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        tags.add_tag_option_to_parser_for_set(parser, 'test')

        parsed_args = parser.parse_args(['--tag', 'tag1'])
        actual = getattr(parsed_args, 'tags', [])
        expected = ['tag1']
        self.assertCountEqual(expected, actual)

        parsed_args = parser.parse_args(['--no-tag'])
        actual = getattr(parsed_args, 'no-tag', [])
        expected = []
        self.assertCountEqual(expected, actual)

    def test_add_tag_option_to_parser_for_unset(self):
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        tags.add_tag_option_to_parser_for_unset(parser, 'test')

        # Test that --tag and --all-tag are mutually exclusive
        self.assertRaises(
            SystemExit, parser.parse_args, ['--tag', 'tag1', '--all-tag']
        )

        parsed_args = parser.parse_args(['--tag', 'tag1'])
        actual = getattr(parsed_args, 'tags', [])
        expected = ['tag1']
        self.assertCountEqual(expected, actual)

        parsed_args = parser.parse_args(['--all-tag'])
        actual = getattr(parsed_args, 'all-tag', [])
        expected = []
        self.assertCountEqual(expected, actual)

    def test_update_tags_for_set(self):
        mock_client = mock.MagicMock()
        mock_obj = mock.MagicMock()
        mock_parsed_args = mock.MagicMock()

        # no-tag True path
        mock_parsed_args.no_tag = True
        mock_parsed_args.tags = ['tag1']
        mock_obj.tags = None
        tags.update_tags_for_set(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_called_once_with(
            mock_obj, list(mock_parsed_args.tags)
        )

        # no-tag False path
        mock_client.set_tags.reset_mock()
        mock_parsed_args.no_tag = False
        mock_parsed_args.tags = ['tag1']
        mock_obj.tags = ['tag2']
        expected_list = ['tag1', 'tag2']
        tags.update_tags_for_set(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_called_once_with(mock_obj, expected_list)

        # no new tags path
        mock_client.set_tags.reset_mock()
        mock_parsed_args.no_tag = False
        mock_parsed_args.tags = None
        mock_obj.tags = ['tag2']
        tags.update_tags_for_set(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_not_called()

    def test_update_tags_for_unset(self):
        mock_client = mock.MagicMock()
        mock_obj = mock.MagicMock()
        mock_parsed_args = mock.MagicMock()

        # No new tags
        mock_obj.tags = ['tag1']
        mock_parsed_args.all_tag = False
        mock_parsed_args.tags = None
        tags.update_tags_for_unset(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_not_called()

        # Clear all tags
        mock_obj.tags = ['tag1']
        mock_parsed_args.all_tag = True
        mock_parsed_args.tags = None
        tags.update_tags_for_unset(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_called_once_with(mock_obj, [])

        # Remove one tag
        mock_client.set_tags.reset_mock()
        mock_obj.tags = ['tag1', 'tag2']
        mock_parsed_args.all_tag = False
        mock_parsed_args.tags = ['tag2']
        tags.update_tags_for_unset(mock_client, mock_obj, mock_parsed_args)
        mock_client.set_tags.assert_called_once_with(mock_obj, ['tag1'])


class TestTagHelps(test_utils.TestCase):
    def _test_tag_method_help(self, meth, exp_normal, exp_enhanced):
        """Vet the help text of the options added by the tag filtering helpers.

        :param meth: One of the ``add_tag_*`` methods.
        :param exp_normal: Expected help output without ``enhance_help``.
        :param exp_enhanced: Expected output with ``enhance_help`` set to
            ``help_enhancer``
        """
        if sys.version_info >= (3, 10):
            options_name = 'options'
        else:
            options_name = 'optional arguments'
        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        meth(parser, 'test')
        self.assertEqual(exp_normal % options_name, parser.format_help())

        parser = argparse.ArgumentParser(prog='run.py', 
            formatter_class=functools.partial(argparse.HelpFormatter, width=78)
        )
        meth(parser, 'test', enhance_help=help_enhancer)
        self.assertEqual(exp_enhanced % options_name, parser.format_help())

    def test_add_tag_filtering_option_to_parser(self):
        self._test_tag_method_help(
            tags.add_tag_filtering_option_to_parser,
            """\
usage: run.py [-h] [--tags <tag>[,<tag>,...]] [--any-tags <tag>[,<tag>,...]]
              [--not-tags <tag>[,<tag>,...]]
              [--not-any-tags <tag>[,<tag>,...]]

%s:
  -h, --help            show this help message and exit
  --tags <tag>[,<tag>,...]
                        List test which have all given tag(s) (Comma-separated
                        list of tags)
  --any-tags <tag>[,<tag>,...]
                        List test which have any given tag(s) (Comma-separated
                        list of tags)
  --not-tags <tag>[,<tag>,...]
                        Exclude test which have all given tag(s) (Comma-
                        separated list of tags)
  --not-any-tags <tag>[,<tag>,...]
                        Exclude test which have any given tag(s) (Comma-
                        separated list of tags)
""",
            """\
usage: run.py [-h] [--tags <tag>[,<tag>,...]] [--any-tags <tag>[,<tag>,...]]
              [--not-tags <tag>[,<tag>,...]]
              [--not-any-tags <tag>[,<tag>,...]]

%s:
  -h, --help            show this help message and exit
  --tags <tag>[,<tag>,...]
                        )sgat fo tsil detarapes-ammoC( )s(gat nevig lla evah
                        hcihw tset tsiL
  --any-tags <tag>[,<tag>,...]
                        )sgat fo tsil detarapes-ammoC( )s(gat nevig yna evah
                        hcihw tset tsiL
  --not-tags <tag>[,<tag>,...]
                        )sgat fo tsil detarapes-ammoC( )s(gat nevig lla evah
                        hcihw tset edulcxE
  --not-any-tags <tag>[,<tag>,...]
                        )sgat fo tsil detarapes-ammoC( )s(gat nevig yna evah
                        hcihw tset edulcxE
""",
        )

    def test_add_tag_option_to_parser_for_create(self):
        self._test_tag_method_help(
            tags.add_tag_option_to_parser_for_create,
            """\
usage: run.py [-h] [--tag <tag> | --no-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  Tag to be added to the test (repeat option to set multiple
               tags)
  --no-tag     No tags associated with the test
""",
            """\
usage: run.py [-h] [--tag <tag> | --no-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  )sgat elpitlum tes ot noitpo taeper( tset eht ot dedda eb ot
               gaT
  --no-tag     tset eht htiw detaicossa sgat oN
""",
        )

    def test_add_tag_option_to_parser_for_set(self):
        self._test_tag_method_help(
            tags.add_tag_option_to_parser_for_set,
            """\
usage: run.py [-h] [--tag <tag>] [--no-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  Tag to be added to the test (repeat option to set multiple
               tags)
  --no-tag     Clear tags associated with the test. Specify both --tag and
               --no-tag to overwrite current tags
""",
            """\
usage: run.py [-h] [--tag <tag>] [--no-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  )sgat elpitlum tes ot noitpo taeper( tset eht ot dedda eb ot
               gaT
  --no-tag     sgat tnerruc etirwrevo ot gat-on-- dna gat-- htob yficepS .tset
               eht htiw detaicossa sgat raelC
""",
        )

    def test_add_tag_option_to_parser_for_unset(self):
        self._test_tag_method_help(
            tags.add_tag_option_to_parser_for_unset,
            """\
usage: run.py [-h] [--tag <tag> | --all-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  Tag to be removed from the test (repeat option to remove
               multiple tags)
  --all-tag    Clear all tags associated with the test
""",
            """\
usage: run.py [-h] [--tag <tag> | --all-tag]

%s:
  -h, --help   show this help message and exit
  --tag <tag>  )sgat elpitlum evomer ot noitpo taeper( tset eht morf devomer
               eb ot gaT
  --all-tag    tset eht htiw detaicossa sgat lla raelC
""",
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Base API Library Tests"""

from keystoneauth1 import exceptions as ksa_exceptions
from keystoneauth1 import session

from osc_lib.api import api
from osc_lib import exceptions
from osc_lib.tests.api import fakes as api_fakes


class TestBaseAPIDefault(api_fakes.TestSession):
    def setUp(self):
        super().setUp()
        self.api = api.BaseAPI()

    def test_baseapi_request_no_url(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        self.assertRaises(
            ksa_exceptions.EndpointNotFound,
            self.api._request,
            'GET',
            '',
        )
        self.assertIsNotNone(self.api.session)
        self.assertNotEqual(self.sess, self.api.session)

    def test_baseapi_request_url(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        ret = self.api._request('GET', self.BASE_URL + '/qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())
        self.assertIsNotNone(self.api.session)
        self.assertNotEqual(self.sess, self.api.session)

    def test_baseapi_request_url_path(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        self.assertRaises(
            ksa_exceptions.EndpointNotFound,
            self.api._request,
            'GET',
            '/qaz',
        )
        self.assertIsNotNone(self.api.session)
        self.assertNotEqual(self.sess, self.api.session)

    def test_baseapi_request_session(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        ret = self.api._request(
            'GET',
            self.BASE_URL + '/qaz',
            session=self.sess,
        )
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())
        self.assertIsNotNone(self.api.session)
        self.assertNotEqual(self.sess, self.api.session)


class TestBaseAPIEndpointArg(api_fakes.TestSession):
    def test_baseapi_endpoint_no_endpoint(self):
        x_api = api.BaseAPI(
            session=self.sess,
        )
        self.assertIsNotNone(x_api.session)
        self.assertEqual(self.sess, x_api.session)
        self.assertIsNone(x_api.endpoint)

        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )

        # Normal url
        self.assertRaises(
            ksa_exceptions.EndpointNotFound,
            x_api._request,
            'GET',
            '/qaz',
        )

        # No leading '/' url
        self.assertRaises(
            ksa_exceptions.EndpointNotFound,
            x_api._request,
            'GET',
            'qaz',
        )

        # Extra leading '/' url
        self.assertRaises(
            ksa_exceptions.connection.UnknownConnectionError,
            x_api._request,
            'GET',
            '//qaz',
        )

    def test_baseapi_endpoint_no_extra(self):
        x_api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )
        self.assertIsNotNone(x_api.session)
        self.assertEqual(self.sess, x_api.session)
        self.assertEqual(self.BASE_URL, x_api.endpoint)

        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )

        # Normal url
        ret = x_api._request('GET', '/qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())

        # No leading '/' url
        ret = x_api._request('GET', 'qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())

        # Extra leading '/' url
        ret = x_api._request('GET', '//qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())

    def test_baseapi_endpoint_extra(self):
        x_api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL + '/',
        )
        self.assertIsNotNone(x_api.session)
        self.assertEqual(self.sess, x_api.session)
        self.assertEqual(self.BASE_URL, x_api.endpoint)

        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )

        # Normal url
        ret = x_api._request('GET', '/qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())

        # No leading '/' url
        ret = x_api._request('GET', 'qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())

        # Extra leading '/' url
        ret = x_api._request('GET', '//qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())


class TestBaseAPIArgs(api_fakes.TestSession):
    def setUp(self):
        super().setUp()
        self.api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )

    def test_baseapi_request_url_path(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        ret = self.api._request('GET', '/qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())
        self.assertIsNotNone(self.api.session)
        self.assertEqual(self.sess, self.api.session)

    def test_baseapi_request_session(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=200,
        )
        new_session = session.Session()
        ret = self.api._request('GET', '/qaz', session=new_session)
        self.assertEqual(api_fakes.RESP_ITEM_1, ret.json())
        self.assertIsNotNone(self.api.session)
        self.assertNotEqual(new_session, self.api.session)


class TestBaseAPICreate(api_fakes.TestSession):
    def setUp(self):
        super().setUp()
        self.api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )

    def test_baseapi_create_post(self):
        self.requests_mock.register_uri(
            'POST',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=202,
        )
        ret = self.api.create('qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

    def test_baseapi_create_put(self):
        self.requests_mock.register_uri(
            'PUT',
            self.BASE_URL + '/qaz',
            json=api_fakes.RESP_ITEM_1,
            status_code=202,
        )
        ret = self.api.create('qaz', method='PUT')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

    def test_baseapi_delete(self):
        self.requests_mock.register_uri(
            'DELETE',
            self.BASE_URL + '/qaz',
            status_code=204,
        )
        ret = self.api.delete('qaz')
        self.assertEqual(204, ret.status_code)


class TestBaseAPIFind(api_fakes.TestSession):
    def setUp(self):
        super().setUp()
        self.api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )

    def test_baseapi_find(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz/1',
            json={'qaz': api_fakes.RESP_ITEM_1},
            status_code=200,
        )
        ret = self.api.find('qaz', '1')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz/1',
            status_code=404,
        )
        self.assertRaises(exceptions.NotFound, self.api.find, 'qaz', '1')

    def test_baseapi_find_attr_by_id(self):
        # All first requests (by name) will fail in this test
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?name=1',
            json={'qaz': []},
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?id=1',
            json={'qaz': [api_fakes.RESP_ITEM_1]},
            status_code=200,
        )
        ret = self.api.find_attr('qaz', '1')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

        # value not found
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?name=0',
            json={'qaz': []},
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?id=0',
            json={'qaz': []},
            status_code=200,
        )
        self.assertRaises(
            exceptions.CommandError,
            self.api.find_attr,
            'qaz',
            '0',
        )

        # Attribute other than 'name'
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?status=UP',
            json={'qaz': [api_fakes.RESP_ITEM_1]},
            status_code=200,
        )
        ret = self.api.find_attr('qaz', 'UP', attr='status')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)
        ret = self.api.find_attr('qaz', value='UP', attr='status')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

    def test_baseapi_find_attr_by_name(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?name=alpha',
            json={'qaz': [api_fakes.RESP_ITEM_1]},
            status_code=200,
        )
        ret = self.api.find_attr('qaz', 'alpha')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

        # value not found
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?name=0',
            json={'qaz': []},
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?id=0',
            json={'qaz': []},
            status_code=200,
        )
        self.assertRaises(
            exceptions.CommandError,
            self.api.find_attr,
            'qaz',
            '0',
        )

        # Attribute other than 'name'
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?status=UP',
            json={'qaz': [api_fakes.RESP_ITEM_1]},
            status_code=200,
        )
        ret = self.api.find_attr('qaz', 'UP', attr='status')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)
        ret = self.api.find_attr('qaz', value='UP', attr='status')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

    def test_baseapi_find_attr_path_resource(self):
        # Test resource different than path
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/wsx?name=1',
            json={'qaz': []},
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/wsx?id=1',
            json={'qaz': [api_fakes.RESP_ITEM_1]},
            status_code=200,
        )
        ret = self.api.find_attr('wsx', '1', resource='qaz')
        self.assertEqual(api_fakes.RESP_ITEM_1, ret)

    def test_baseapi_find_bulk_none(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.find_bulk('qaz')
        self.assertEqual(api_fakes.LIST_RESP, ret)
        # Verify headers arg does not interfere
        ret = self.api.find_bulk('qaz', headers={})
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_find_bulk_one(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.find_bulk('qaz', id='1')
        self.assertEqual([api_fakes.LIST_RESP[0]], ret)
        # Verify headers arg does not interfere with search
        ret = self.api.find_bulk('qaz', id='1', headers={})
        self.assertEqual([api_fakes.LIST_RESP[0]], ret)

        ret = self.api.find_bulk('qaz', id='0')
        self.assertEqual([], ret)

        ret = self.api.find_bulk('qaz', name='beta')
        self.assertEqual([api_fakes.LIST_RESP[1]], ret)

        ret = self.api.find_bulk('qaz', error='bogus')
        self.assertEqual([], ret)

    def test_baseapi_find_bulk_two(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.find_bulk('qaz', id='1', name='alpha')
        self.assertEqual([api_fakes.LIST_RESP[0]], ret)

        ret = self.api.find_bulk('qaz', id='1', name='beta')
        self.assertEqual([], ret)

        ret = self.api.find_bulk('qaz', id='1', error='beta')
        self.assertEqual([], ret)

    def test_baseapi_find_bulk_dict(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json={'qaz': api_fakes.LIST_RESP},
            status_code=200,
        )
        ret = self.api.find_bulk('qaz', id='1')
        self.assertEqual([api_fakes.LIST_RESP[0]], ret)


class TestBaseAPIList(api_fakes.TestSession):
    def setUp(self):
        super().setUp()
        self.api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )

    def test_baseapi_list_no_args(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz',
            json=api_fakes.LIST_RESP,
            status_code=204,
        )
        ret = self.api.list('/qaz')
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_list_params(self):
        params = {'format': 'json'}
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '?format=json',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.list('', **params)
        self.assertEqual(api_fakes.LIST_RESP, ret)

        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?format=json',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.list('qaz', **params)
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_list_body(self):
        self.requests_mock.register_uri(
            'POST',
            self.BASE_URL + '/qaz',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.list('qaz', body=api_fakes.LIST_BODY)
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_list_detailed(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz/details',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.list('qaz', detailed=True)
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_list_filtered(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?attr=value',
            json=api_fakes.LIST_RESP,
            status_code=200,
        )
        ret = self.api.list('qaz', attr='value')
        self.assertEqual(api_fakes.LIST_RESP, ret)

    def test_baseapi_list_wrapped(self):
        self.requests_mock.register_uri(
            'GET',
            self.BASE_URL + '/qaz?attr=value',
            json={'responses': api_fakes.LIST_RESP},
            status_code=200,
        )
        ret = self.api.list('qaz', attr='value')
        self.assertEqual({'responses': api_fakes.LIST_RESP}, ret)



#   Copyright 2016 NEC Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import abc
import argparse
import logging
import typing as ty

from cliff import command
from cliff import lister
from cliff import show

from osc_lib import exceptions
from osc_lib.i18n import _


class CommandMeta(abc.ABCMeta):
    def __new__(
        mcs: ty.Type['CommandMeta'],
        name: str,
        bases: ty.Tuple[ty.Type[ty.Any], ...],
        namespace: ty.Dict[str, ty.Any],
    ) -> 'CommandMeta':
        if 'log' not in namespace:
            namespace['log'] = logging.getLogger(
                namespace['__module__'] + '.' + name
            )
        return super().__new__(mcs, name, bases, namespace)


class Command(command.Command, metaclass=CommandMeta):
    log: logging.Logger

    def run(self, parsed_args: argparse.Namespace) -> int:
        self.log.debug('run(%s)', parsed_args)
        return super().run(parsed_args)

    def validate_os_beta_command_enabled(self) -> None:
        if not self.app.options.os_beta_command:
            msg = _(
                'Caution: This is a beta command and subject to '
                'change. Use global option --os-beta-command '
                'to enable this command.'
            )
            raise exceptions.CommandError(msg)

    def deprecated_option_warning(
        self, old_option: str, new_option: str
    ) -> None:
        """Emit a warning for use of a deprecated option"""
        self.log.warning(
            _("The %(old)s option is deprecated, please use %(new)s instead.")
            % {'old': old_option, 'new': new_option}
        )


class Lister(Command, lister.Lister):
    pass


class ShowOne(Command, show.ShowOne):
    pass



#   Copyright 2012-2013 OpenStack Foundation
#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
import json as jsonutils
import os
from unittest import mock

from cliff import columns as cliff_columns
import fixtures
from keystoneauth1 import loading
from openstack.config import cloud_region
from openstack.config import defaults
from oslo_utils import importutils
from requests_mock.contrib import fixture
import testtools

from osc_lib import clientmanager
from osc_lib import shell
from osc_lib.tests import fakes


def fake_execute(shell, cmd):
    """Pretend to execute shell commands."""
    return shell.run(cmd.split())


def make_shell(shell_class=None):
    """Create a new command shell and mock out some bits."""
    if shell_class is None:
        shell_class = shell.OpenStackShell
    _shell = shell_class()
    _shell.command_manager = mock.Mock()
    # _shell.cloud = mock.Mock()

    return _shell


def opt2attr(opt):
    if opt.startswith('--os-'):
        attr = opt[5:]
    elif opt.startswith('--'):
        attr = opt[2:]
    else:
        attr = opt
    return attr.lower().replace('-', '_')


def opt2env(opt):
    return opt[2:].upper().replace('-', '_')


class EnvFixture(fixtures.Fixture):
    """Environment Fixture.

    This fixture replaces os.environ with provided env or an empty env.
    """

    def __init__(self, env=None):
        self.new_env = env or {}

    def _setUp(self):
        self.orig_env, os.environ = os.environ, self.new_env
        self.addCleanup(self.revert)

    def revert(self):
        os.environ = self.orig_env


class ParserException(Exception):
    pass


class TestCase(testtools.TestCase):
    def setUp(self):
        testtools.TestCase.setUp(self)

        if (
            os.environ.get("OS_STDOUT_CAPTURE") == "True"
            or os.environ.get("OS_STDOUT_CAPTURE") == "1"
        ):
            stdout = self.useFixture(fixtures.StringStream("stdout")).stream
            self.useFixture(fixtures.MonkeyPatch("sys.stdout", stdout))

        if (
            os.environ.get("OS_STDERR_CAPTURE") == "True"
            or os.environ.get("OS_STDERR_CAPTURE") == "1"
        ):
            stderr = self.useFixture(fixtures.StringStream("stderr")).stream
            self.useFixture(fixtures.MonkeyPatch("sys.stderr", stderr))

    def assertNotCalled(self, m, msg=None):
        """Assert a function was not called"""

        if m.called:
            if not msg:
                msg = f'method {m} should not have been called'
            self.fail(msg)


class TestCommand(TestCase):
    """Test cliff command classes"""

    def setUp(self):
        super().setUp()
        # Build up a fake app
        self.fake_stdout = fakes.FakeStdout()
        self.fake_log = fakes.FakeLog()
        self.app = fakes.FakeApp(self.fake_stdout, self.fake_log)
        self.app.client_manager = fakes.FakeClientManager()

    def check_parser(self, cmd, args, verify_args):
        cmd_parser = cmd.get_parser('check_parser')
        try:
            parsed_args = cmd_parser.parse_args(args)
        except SystemExit:
            raise ParserException("Argument parse failed")
        for av in verify_args:
            attr, value = av
            if attr:
                self.assertIn(attr, parsed_args)
                self.assertEqual(value, getattr(parsed_args, attr))
        return parsed_args

    def assertItemEqual(self, expected, actual):
        """Compare item considering formattable columns.

        This method compares an observed item to an expected item column by
        column. If a column is a formattable column, observed and expected
        columns are compared using human_readable() and machine_readable().
        """
        self.assertEqual(len(expected), len(actual))
        for col_expected, col_actual in zip(expected, actual):
            if isinstance(col_expected, cliff_columns.FormattableColumn):
                self.assertIsInstance(col_actual, col_expected.__class__)
                self.assertEqual(
                    col_expected.human_readable(), col_actual.human_readable()
                )
                self.assertEqual(
                    col_expected.machine_readable(),
                    col_actual.machine_readable(),
                )
            else:
                self.assertEqual(col_expected, col_actual)

    def assertListItemEqual(self, expected, actual):
        """Compare a list of items considering formattable columns.

        Each pair of observed and expected items are compared
        using assertItemEqual() method.
        """
        self.assertEqual(len(expected), len(actual))
        for item_expected, item_actual in zip(expected, actual):
            self.assertItemEqual(item_expected, item_actual)


class TestClientManager(TestCase):
    """ClientManager class test framework"""

    default_password_auth = {
        'auth_url': fakes.AUTH_URL,
        'username': fakes.USERNAME,
        'password': fakes.PASSWORD,
        'project_name': fakes.PROJECT_NAME,
    }
    default_token_auth = {
        'auth_url': fakes.AUTH_URL,
        'token': fakes.AUTH_TOKEN,
    }

    def setUp(self):
        super().setUp()
        self.mock = mock.Mock()
        self.requests = self.useFixture(fixture.Fixture())
        # fake v2password token retrieval
        self.stub_auth(json=fakes.TEST_RESPONSE_DICT)
        # fake token and token_endpoint retrieval
        self.stub_auth(
            json=fakes.TEST_RESPONSE_DICT,
            url='/'.join([fakes.AUTH_URL, 'v2.0/tokens']),
        )
        # fake v3password token retrieval
        self.stub_auth(
            json=fakes.TEST_RESPONSE_DICT_V3,
            url='/'.join([fakes.AUTH_URL, 'v3/auth/tokens']),
        )
        # fake password token retrieval
        self.stub_auth(
            json=fakes.TEST_RESPONSE_DICT_V3,
            url='/'.join([fakes.AUTH_URL, 'auth/tokens']),
        )
        # fake password version endpoint discovery
        self.stub_auth(
            json=fakes.TEST_VERSIONS, url=fakes.AUTH_URL, verb='GET'
        )

        # Mock the auth plugin
        self.auth_mock = mock.Mock()

    def stub_auth(self, json=None, url=None, verb=None, **kwargs):
        subject_token = fakes.AUTH_TOKEN
        base_url = fakes.AUTH_URL
        if json:
            text = jsonutils.dumps(json)
            headers = {
                'X-Subject-Token': subject_token,
                'Content-Type': 'application/json',
            }
        if not url:
            url = '/'.join([base_url, 'tokens'])
        url = url.replace("/?", "?")
        if not verb:
            verb = 'POST'
        self.requests.register_uri(
            verb,
            url,
            headers=headers,
            text=text,
        )

    def _clientmanager_class(self):
        """Allow subclasses to override the ClientManager class"""
        return clientmanager.ClientManager

    def _make_clientmanager(
        self,
        auth_args=None,
        config_args=None,
        identity_api_version=None,
        auth_plugin_name=None,
        auth_required=None,
    ):
        if identity_api_version is None:
            identity_api_version = '2.0'
        if auth_plugin_name is None:
            auth_plugin_name = 'password'

        if auth_plugin_name.endswith('password'):
            auth_dict = copy.deepcopy(self.default_password_auth)
        elif auth_plugin_name.endswith('token'):
            auth_dict = copy.deepcopy(self.default_token_auth)
        else:
            auth_dict = {}

        if auth_args is not None:
            auth_dict = auth_args

        cli_options = defaults.get_defaults()
        cli_options.update(
            {
                'auth_type': auth_plugin_name,
                'auth': auth_dict,
                'interface': fakes.INTERFACE,
                'region_name': fakes.REGION_NAME,
                # 'workflow_api_version': '2',
            }
        )
        if config_args is not None:
            cli_options.update(config_args)

        loader = loading.get_plugin_loader(auth_plugin_name)
        auth_plugin = loader.load_from_options(**auth_dict)
        client_manager = self._clientmanager_class()(
            cli_options=cloud_region.CloudRegion(
                name='t1',
                region_name='1',
                config=cli_options,
                auth_plugin=auth_plugin,
            ),
            api_version={
                'identity': identity_api_version,
            },
        )
        client_manager._auth_required = auth_required is True
        client_manager.setup_auth()
        client_manager.auth_ref

        self.assertEqual(
            auth_plugin_name,
            client_manager.auth_plugin_name,
        )
        return client_manager


class TestShell(TestCase):
    # Full name of the OpenStackShell class to test (cliff.app.App subclass)
    shell_class_name = "osc_lib.shell.OpenStackShell"

    def setUp(self):
        super().setUp()
        self.shell_class = importutils.import_class(self.shell_class_name)
        self.cmd_patch = mock.patch(self.shell_class_name + ".run_subcommand")
        self.cmd_save = self.cmd_patch.start()
        self.addCleanup(self.cmd_patch.stop)
        self.app = mock.Mock("Test Shell")

    def _assert_initialize_app_arg(self, cmd_options, default_args):
        """Check the args passed to initialize_app()

        The argv argument to initialize_app() is the remainder from parsing
        global options declared in both cliff.app and
        osc_lib.OpenStackShell build_option_parser().  Any global
        options passed on the command line should not be in argv but in
        _shell.options.
        """

        with mock.patch(
            self.shell_class_name + ".initialize_app",
            self.app,
        ):
            _shell = make_shell(shell_class=self.shell_class)
            _cmd = cmd_options + " module list"
            fake_execute(_shell, _cmd)

            self.app.assert_called_with(["module", "list"])
            for k in default_args.keys():
                self.assertEqual(
                    default_args[k],
                    vars(_shell.options)[k],
                    f"{k} does not match",
                )

    def _assert_cloud_region_arg(self, cmd_options, default_args):
        """Check the args passed to OpenStackConfig.get_one()

        The argparse argument to get_one() is an argparse.Namespace
        object that contains all of the options processed to this point in
        initialize_app().
        """

        cloud = mock.Mock(name="cloudy")
        cloud.config = {}
        self.occ_get_one = mock.Mock(return_value=cloud)
        with mock.patch(
            "openstack.config.loader.OpenStackConfig.get_one",
            self.occ_get_one,
        ):
            _shell = make_shell(shell_class=self.shell_class)
            _cmd = cmd_options + " module list"
            fake_execute(_shell, _cmd)

            self.app.assert_called_with(["module", "list"])
            opts = self.occ_get_one.call_args[1]['argparse']
            for k in default_args.keys():
                self.assertEqual(
                    default_args[k],
                    vars(opts)[k],
                    f"{k} does not match",
                )

    def _test_options_init_app(self, test_opts):
        """Test options on the command line"""
        for opt in test_opts.keys():
            if not test_opts[opt][1]:
                continue
            key = opt2attr(opt)
            if isinstance(test_opts[opt][0], str):
                cmd = opt + " " + test_opts[opt][0]
            else:
                cmd = opt
            kwargs = {
                key: test_opts[opt][0],
            }
            self._assert_initialize_app_arg(cmd, kwargs)

    def _test_env_init_app(self, test_opts):
        """Test options in the environment"""
        for opt in test_opts.keys():
            if not test_opts[opt][2]:
                continue
            key = opt2attr(opt)
            kwargs = {
                key: test_opts[opt][0],
            }
            env = {
                opt2env(opt): test_opts[opt][0],
            }
            os.environ = env.copy()
            self._assert_initialize_app_arg("", kwargs)

    def _test_options_get_one_cloud(self, test_opts):
        """Test options sent "to openstack.config"""
        for opt in test_opts.keys():
            if not test_opts[opt][1]:
                continue
            key = opt2attr(opt)
            if isinstance(test_opts[opt][0], str):
                cmd = opt + " " + test_opts[opt][0]
            else:
                cmd = opt
            kwargs = {
                key: test_opts[opt][0],
            }
            self._assert_cloud_region_arg(cmd, kwargs)

    def _test_env_get_one_cloud(self, test_opts):
        """Test environment options sent "to openstack.config"""
        for opt in test_opts.keys():
            if not test_opts[opt][2]:
                continue
            key = opt2attr(opt)
            kwargs = {
                key: test_opts[opt][0],
            }
            env = {
                opt2env(opt): test_opts[opt][0],
            }
            os.environ = env.copy()
            self._assert_cloud_region_arg("", kwargs)



#   Copyright 2012-2013 OpenStack Foundation
#   Copyright 2015 Dean Troyer
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Command-line interface to the OpenStack APIs"""

import getpass
import logging
import sys
import traceback
import typing as ty

from cliff import _argparse
from cliff import app
from cliff import command
from cliff import commandmanager
from cliff import complete
from cliff import help
from cliff import interactive
from oslo_utils import importutils
from oslo_utils import strutils

from osc_lib.cli import client_config as cloud_config
from osc_lib import clientmanager
from osc_lib.command import timing
from osc_lib import exceptions as exc
from osc_lib.i18n import _
from osc_lib import logs
from osc_lib import utils
from osc_lib import version

osprofiler_profiler = importutils.try_import("osprofiler.profiler")


DEFAULT_DOMAIN = 'default'
DEFAULT_INTERFACE = 'public'


def prompt_for_password(prompt: ty.Optional[str] = None) -> str:
    """Prompt user for a password

    Prompt for a password if stdin is a tty.
    """

    if not prompt:
        prompt = 'Password: '
    pw = None
    # If stdin is a tty, try prompting for the password
    if hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
        # Check for Ctl-D
        try:
            pw = getpass.getpass(prompt)
        except EOFError:
            pass
    # No password because we did't have a tty or nothing was entered
    if not pw:
        raise exc.CommandError(
            _(
                "No password entered, or found via"
                " --os-password or OS_PASSWORD"
            ),
        )

    return pw


class OpenStackShell(app.App):
    CONSOLE_MESSAGE_FORMAT = '%(levelname)s: %(name)s %(message)s'

    client_manager: clientmanager.ClientManager

    log = logging.getLogger(__name__)
    timing_data: ty.List[ty.Any] = []

    def __init__(
        self,
        description: ty.Optional[str] = None,
        version: ty.Optional[str] = None,
        command_manager: ty.Optional[commandmanager.CommandManager] = None,
        stdin: ty.Optional[ty.TextIO] = None,
        stdout: ty.Optional[ty.TextIO] = None,
        stderr: ty.Optional[ty.TextIO] = None,
        interactive_app_factory: ty.Optional[
            ty.Type['interactive.InteractiveApp']
        ] = None,
        deferred_help: bool = False,
    ) -> None:
        # Patch command.Command to add a default auth_required = True
        setattr(command.Command, 'auth_required', True)

        # Some commands do not need authentication
        setattr(help.HelpCommand, 'auth_required', False)
        setattr(complete.CompleteCommand, 'auth_required', False)

        # Slight change to the meaning of --debug
        self.DEFAULT_DEBUG_VALUE = None
        self.DEFAULT_DEBUG_HELP = 'Set debug logging and traceback on errors.'

        # Do default for positionals
        if not command_manager:
            cm = commandmanager.CommandManager('openstack.cli')
        else:
            cm = command_manager

        super().__init__(
            description=__doc__.strip(),
            version=version,
            command_manager=cm,
            deferred_help=True,
        )

        # Until we have command line arguments parsed, dump any stack traces
        self.dump_stack_trace = True

        # Set in subclasses
        self.api_version = None

        self.command_options: ty.List[str] = []

        self.do_profile = False

    def configure_logging(self) -> None:
        """Configure logging for the app."""
        self.log_configurator = logs.LogConfigurator(self.options)
        self.dump_stack_trace = self.log_configurator.dump_trace

    def run(self, argv: ty.List[str]) -> int:
        ret_val = 1
        self.command_options = argv
        try:
            ret_val = super().run(argv)
            return ret_val
        except Exception as e:
            if not logging.getLogger('').handlers:
                logging.basicConfig()
            if self.dump_stack_trace:
                self.log.error(traceback.format_exc())
            else:
                self.log.error('Exception raised: ' + str(e))

            return ret_val

        finally:
            self.log.debug("END return value: %s", ret_val)

    def init_profile(self) -> None:
        self.do_profile = osprofiler_profiler and self.options.profile
        if self.do_profile:
            osprofiler_profiler.init(self.options.profile)

    def close_profile(self) -> None:
        if self.do_profile:
            profiler = osprofiler_profiler.get()
            trace_id = profiler.get_base_id()
            # Short ID for OpenTracing-based driver (64-bit id)
            short_id = profiler.get_shorten_id(trace_id)

            # NOTE(dbelova): let's use warning log level to see these messages
            # printed. In fact we can define custom log level here with value
            # bigger than most big default one (CRITICAL) or something like
            # that (PROFILE = 60 for instance), but not sure we need it here.
            self.log.warning(f"Trace ID: {trace_id}")
            self.log.warning(
                f"Short trace ID for OpenTracing-based drivers: {short_id}"
            )
            self.log.warning(
                "Display trace data with command:\n"
                f"osprofiler trace show --html {trace_id} "
            )

    def run_subcommand(self, argv: ty.List[str]) -> int:
        self.init_profile()
        try:
            ret_value = super().run_subcommand(argv)
        finally:
            self.close_profile()
        return ret_value

    def interact(self) -> None:
        self.init_profile()
        try:
            ret_value = super().interact()
        finally:
            self.close_profile()
        return ret_value

    def build_option_parser(
        self,
        description: ty.Optional[str],
        version: ty.Optional[str],
        argparse_kwargs: ty.Optional[ty.Dict[str, ty.Any]] = None,
    ) -> _argparse.ArgumentParser:
        parser = super().build_option_parser(
            description,
            version,
            argparse_kwargs,
        )

        # service token auth argument
        parser.add_argument(
            '--os-cloud',
            metavar='<cloud-config-name>',
            dest='cloud',
            default=utils.env('OS_CLOUD'),
            help=_('Cloud name in clouds.yaml (Env: OS_CLOUD)'),
        )
        # Global arguments
        parser.add_argument(
            '--os-region-name',
            metavar='<auth-region-name>',
            dest='region_name',
            default=utils.env('OS_REGION_NAME'),
            help=_('Authentication region name (Env: OS_REGION_NAME)'),
        )
        parser.add_argument(
            '--os-cacert',
            metavar='<ca-bundle-file>',
            dest='cacert',
            default=utils.env('OS_CACERT', default=None),
            help=_('CA certificate bundle file (Env: OS_CACERT)'),
        )
        parser.add_argument(
            '--os-cert',
            metavar='<certificate-file>',
            dest='cert',
            default=utils.env('OS_CERT'),
            help=_('Client certificate bundle file (Env: OS_CERT)'),
        )
        parser.add_argument(
            '--os-key',
            metavar='<key-file>',
            dest='key',
            default=utils.env('OS_KEY'),
            help=_('Client certificate key file (Env: OS_KEY)'),
        )
        verify_group = parser.add_mutually_exclusive_group()
        verify_group.add_argument(
            '--verify',
            action='store_true',
            default=None,
            help=_('Verify server certificate (default)'),
        )
        verify_group.add_argument(
            '--insecure',
            action='store_true',
            default=None,
            help=_('Disable server certificate verification'),
        )
        parser.add_argument(
            '--os-default-domain',
            metavar='<auth-domain>',
            dest='default_domain',
            default=utils.env('OS_DEFAULT_DOMAIN', default=DEFAULT_DOMAIN),
            help=_('Default domain ID, default=%s. (Env: OS_DEFAULT_DOMAIN)')
            % DEFAULT_DOMAIN,
        )
        parser.add_argument(
            '--os-interface',
            metavar='<interface>',
            dest='interface',
            choices=['admin', 'public', 'internal'],
            # do not set default value inside utils.env() else user's cloud
            # config key 'interface' will be ignored. Use OSC_Config's ctor
            # option 'override_defaults' below instead.
            default=utils.env('OS_INTERFACE'),
            help=_(
                'Select an interface type.'
                ' Valid interface types: [admin, public, internal].'
                ' default=%s, (Env: OS_INTERFACE)'
            )
            % DEFAULT_INTERFACE,
        )
        parser.add_argument(
            '--os-service-provider',
            metavar='<service_provider>',
            dest='service_provider',
            default=utils.env('OS_SERVICE_PROVIDER'),
            help=_(
                'Authenticate with and perform the command on a service'
                ' provider using Keystone-to-keystone federation. Must'
                ' also specify the remote project option.'
            ),
        )
        remote_project_group = parser.add_mutually_exclusive_group()
        remote_project_group.add_argument(
            '--os-remote-project-name',
            metavar='<remote_project_name>',
            dest='remote_project_name',
            default=utils.env('OS_REMOTE_PROJECT_NAME'),
            help=_(
                'Project name when authenticating to a service provider'
                ' if using Keystone-to-Keystone federation.'
            ),
        )
        remote_project_group.add_argument(
            '--os-remote-project-id',
            metavar='<remote_project_id>',
            dest='remote_project_id',
            default=utils.env('OS_REMOTE_PROJECT_ID'),
            help=_(
                'Project ID when authenticating to a service provider'
                ' if using Keystone-to-Keystone federation.'
            ),
        )
        remote_project_domain_group = parser.add_mutually_exclusive_group()
        remote_project_domain_group.add_argument(
            '--os-remote-project-domain-name',
            metavar='<remote_project_domain_name>',
            dest='remote_project_domain_name',
            default=utils.env('OS_REMOTE_PROJECT_DOMAIN_NAME'),
            help=_(
                'Domain name of the project when authenticating to a'
                ' service provider if using Keystone-to-Keystone'
                ' federation.'
            ),
        )
        remote_project_domain_group.add_argument(
            '--os-remote-project-domain-id',
            metavar='<remote_project_domain_id>',
            dest='remote_project_domain_id',
            default=utils.env('OS_REMOTE_PROJECT_DOMAIN_ID'),
            help=_(
                'Domain ID of the project when authenticating to a'
                ' service provider if using Keystone-to-Keystone'
                ' federation.'
            ),
        )
        parser.add_argument(
            '--timing',
            default=False,
            action='store_true',
            help=_("Print API call timing info"),
        )
        parser.add_argument(
            '--os-beta-command',
            action='store_true',
            help=_("Enable beta commands which are subject to change"),
        )

        # osprofiler HMAC key argument
        if osprofiler_profiler:
            parser.add_argument(
                '--os-profile',
                metavar='hmac-key',
                dest='profile',
                default=utils.env('OS_PROFILE'),
                help=_('HMAC key for encrypting profiling context data'),
            )

        return parser

    """
    Break up initialize_app() so that overriding it in a subclass does not
    require duplicating a lot of the method

    * super()
    * _final_defaults()
    * OpenStackConfig
    * get_one
    * _load_plugins()
    * _load_commands()
    * ClientManager

    """

    def _final_defaults(self) -> None:
        # Set the default plugin to None
        # NOTE(dtroyer): This is here to set up for setting it to a default
        #                in the calling CLI
        self._auth_type = None

        # Converge project/tenant options
        project_id = getattr(self.options, 'project_id', None)
        project_name = getattr(self.options, 'project_name', None)
        tenant_id = getattr(self.options, 'tenant_id', None)
        tenant_name = getattr(self.options, 'tenant_name', None)

        # handle some v2/v3 authentication inconsistencies by just acting like
        # both the project and tenant information are both present. This can
        # go away if we stop registering all the argparse options together.
        if project_id and not tenant_id:
            self.options.tenant_id = project_id
        if project_name and not tenant_name:
            self.options.tenant_name = project_name
        if tenant_id and not project_id:
            self.options.project_id = tenant_id
        if tenant_name and not project_name:
            self.options.project_name = tenant_name

        # Save default domain
        self.default_domain = self.options.default_domain

    def _load_plugins(self) -> None:
        """Load plugins via stevedore

        osc-lib has no opinion on what plugins should be loaded
        """
        pass

    def _load_commands(self) -> None:
        """Load commands via cliff/stevedore

        osc-lib has no opinion on what commands should be loaded
        """
        pass

    def initialize_app(self, argv: ty.List[str]) -> None:
        """Global app init bits:

        * set up API versions
        * validate authentication info
        * authenticate against Identity if requested
        """

        # Parent __init__ parses argv into self.options
        super().initialize_app(argv)
        self.log.info(
            "START with options: %s",
            strutils.mask_password(" ".join(self.command_options))
            if self.command_options
            else "",
        )
        self.log.debug("options: %s", strutils.mask_password(self.options))

        # Callout for stuff between superclass init and o-c-c
        self._final_defaults()

        # Do configuration file handling
        try:
            self.cloud_config = cloud_config.OSC_Config(
                pw_func=prompt_for_password,
                override_defaults={
                    'interface': DEFAULT_INTERFACE,
                    'auth_type': self._auth_type,
                },
            )
        except OSError as e:
            self.log.critical("Could not read clouds.yaml configuration file")
            self.print_help_if_requested()
            raise e

        # TODO(thowe): Change cliff so the default value for debug
        # can be set to None.
        if not self.options.debug:
            self.options.debug = None

        # NOTE(dtroyer): Need to do this with validate=False to defer the
        #                auth plugin handling to ClientManager.setup_auth()
        self.cloud = self.cloud_config.get_one(
            cloud=self.options.cloud,
            argparse=self.options,
            validate=False,
        )

        self.log_configurator.configure(self.cloud)
        self.dump_stack_trace = self.log_configurator.dump_trace
        self.log.debug("defaults: %s", self.cloud_config.defaults)
        self.log.debug(
            "cloud cfg: %s", strutils.mask_password(self.cloud.config)
        )

        # Callout for stuff between o-c-c and ClientManager
        # self._initialize_app_2(self.options)

        self._load_plugins()

        self._load_commands()

        # Handle deferred help and exit
        self.print_help_if_requested()

        self.client_manager = clientmanager.ClientManager(
            cli_options=self.cloud,
            api_version=self.api_version,
            pw_func=prompt_for_password,
        )

    def prepare_to_run_command(self, cmd: 'command.Command') -> None:
        """Set up auth and API versions"""
        self.log.debug(
            'command: %s -> %s.%s (auth=%s)',
            getattr(cmd, 'cmd_name', '<none>'),
            cmd.__class__.__module__,
            cmd.__class__.__name__,
            getattr(cmd, 'auth_required', None),
        )

        # NOTE(dtroyer): If auth is not required for a command, skip
        #                get_one()'s validation to avoid loading plugins
        validate = getattr(cmd, 'auth_required', False)

        # NOTE(dtroyer): Save the auth required state of the _current_ command
        #                in the ClientManager
        self.client_manager._auth_required = validate

        # Validate auth options
        self.cloud = self.cloud_config.get_one(
            cloud=self.options.cloud,
            argparse=self.options,
            validate=validate,
            app_name=self.client_manager._app_name,
            app_version=self.client_manager._app_version,
            additional_user_agent=[('osc-lib', version.version_string)],
        )
        # Push the updated args into ClientManager
        self.client_manager._cli_options = self.cloud

        if validate:
            self.client_manager.setup_auth()
            if hasattr(cmd, 'required_scope') and cmd.required_scope:
                # let the command decide whether we need a scoped token
                self.client_manager.validate_scope()
            # Trigger the Identity client to initialize
            self.client_manager.session.auth.auth_ref = (  # type: ignore
                self.client_manager.auth_ref
            )
        return

    def clean_up(
        self,
        cmd: 'command.Command',
        result: int,
        err: ty.Optional[BaseException],
    ) -> None:
        self.log.debug('clean_up %s: %s', cmd.__class__.__name__, err or '')

        # Close SDK connection if available to have proper cleanup there
        if getattr(self.client_manager, "sdk_connection", None) is not None:
            self.client_manager.sdk_connection.close()

        # Close session if available
        if getattr(self.client_manager, "session", None) is not None:
            self.client_manager.session.session.close()

        # Process collected timing data
        if self.options.timing:
            # Get session data
            self.timing_data.extend(
                self.client_manager.session.get_timings(),
            )

            # Use the Timing pseudo-command to generate the output
            tcmd = timing.Timing(self, self.options)
            tparser = tcmd.get_parser('Timing')

            # If anything other than prettytable is specified, force csv
            format = 'table'
            # Check the formatter used in the actual command
            if (
                hasattr(cmd, 'formatter')
                and hasattr(cmd, '_formatter_plugins')
                and cmd.formatter != cmd._formatter_plugins['table'].obj
            ):
                format = 'csv'

            sys.stdout.write('\n')
            targs = tparser.parse_args(['-f', format])
            tcmd.run(targs)


def main(argv: ty.Optional[ty.List[str]] = None) -> int:
    if argv is None:
        argv = sys.argv[1:]
    return OpenStackShell().run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Exception definitions."""

import typing as ty


class CommandError(Exception):
    pass


class AuthorizationFailure(Exception):
    pass


class PluginAttributeError(Exception):
    """A plugin threw an AttributeError while being lazily loaded."""

    # This *must not* inherit from AttributeError;
    # that would defeat the whole purpose.
    pass


class NoTokenLookupException(Exception):
    """This does not support looking up endpoints from an existing token."""

    pass


class EndpointNotFound(Exception):
    """Could not find Service or Region in Service Catalog."""

    pass


class UnsupportedVersion(Exception):
    """The user is trying to use an unsupported version of the API"""

    pass


class InvalidValue(Exception):
    """An argument value is not valid: wrong type, out of range, etc"""

    message = "Supplied value is not valid"


class ClientException(Exception):
    """The base exception class for all exceptions this library raises."""

    http_status: int
    message: str

    def __init__(
        self,
        code: ty.Union[int, str],
        message: ty.Optional[str] = None,
        details: ty.Optional[str] = None,
    ):
        if not isinstance(code, int) and message is None:
            message = code
            code = self.http_status
        self.code = code
        self.message = message or self.__class__.message
        self.details = details

    def __str__(self) -> str:
        return f"{self.message} (HTTP {self.code})"


class BadRequest(ClientException):
    """HTTP 400 - Bad request: you sent some malformed data."""

    http_status = 400
    message = "Bad request"


class Unauthorized(ClientException):
    """HTTP 401 - Unauthorized: bad credentials."""

    http_status = 401
    message = "Unauthorized"


class Forbidden(ClientException):
    """HTTP 403 - Forbidden: not authorized to access to this resource."""

    http_status = 403
    message = "Forbidden"


class NotFound(ClientException):
    """HTTP 404 - Not found"""

    http_status = 404
    message = "Not found"


class Conflict(ClientException):
    """HTTP 409 - Conflict"""

    http_status = 409
    message = "Conflict"


class OverLimit(ClientException):
    """HTTP 413 - Over limit: reached the API limits for this time period."""

    http_status = 413
    message = "Over limit"


# NotImplemented is a python keyword.
class HTTPNotImplemented(ClientException):
    """HTTP 501 - Not Implemented: server does not support this operation."""

    http_status = 501
    message = "Not Implemented"


# In Python 2.4 Exception is old-style and thus doesn't have a __subclasses__()
# so we can do this:
#     _code_map = dict((c.http_status, c)
#                      for c in ClientException.__subclasses__())
#
# Instead, we have to hardcode it:
_code_map = dict(
    (c.http_status, c)
    for c in [
        BadRequest,
        Unauthorized,
        Forbidden,
        NotFound,
        OverLimit,
        HTTPNotImplemented,
    ]
)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Application logging"""

import argparse
import logging
import sys
import typing as ty
import warnings

from openstack.config import cloud_config


def get_loggers() -> ty.Dict[str, str]:
    loggers = {}
    for logkey in logging.Logger.manager.loggerDict.keys():
        loggers[logkey] = logging.getLevelName(logging.getLogger(logkey).level)
    return loggers


def log_level_from_options(options: argparse.Namespace) -> int:
    # if --debug, --quiet or --verbose is not specified,
    # the default logging level is warning
    log_level = logging.WARNING
    if options.verbose_level == 0:
        # --quiet
        log_level = logging.ERROR
    elif options.verbose_level == 2:
        # One --verbose
        log_level = logging.INFO
    elif options.verbose_level >= 3:
        # Two or more --verbose
        log_level = logging.DEBUG
    return log_level


def log_level_from_string(level_string: str) -> int:
    log_level = {
        'critical': logging.CRITICAL,
        'error': logging.ERROR,
        'warning': logging.WARNING,
        'info': logging.INFO,
        'debug': logging.DEBUG,
    }.get(level_string, logging.WARNING)
    return log_level


def log_level_from_config(config: ty.Mapping[str, ty.Any]) -> int:
    # Check the command line option
    verbose_level = config.get('verbose_level')
    if config.get('debug', False):
        verbose_level = 3
    if verbose_level == 0:
        verbose_level = 'error'
    elif verbose_level == 1:
        # If a command line option has not been specified, check the
        # configuration file
        verbose_level = config.get('log_level', 'warning')
    elif verbose_level == 2:
        verbose_level = 'info'
    else:
        verbose_level = 'debug'
    return log_level_from_string(verbose_level)


def set_warning_filter(log_level: int) -> None:
    if log_level == logging.ERROR:
        warnings.simplefilter("ignore")
    elif log_level == logging.WARNING:
        warnings.simplefilter("ignore")
    elif log_level == logging.INFO:
        warnings.simplefilter("once")


class _FileFormatter(logging.Formatter):
    """Customize the logging format for logging handler"""

    _LOG_MESSAGE_BEGIN = (
        '%(asctime)s.%(msecs)03d %(process)d %(levelname)s %(name)s '
    )
    _LOG_MESSAGE_CONTEXT = '[%(cloud)s %(username)s %(project)s] '
    _LOG_MESSAGE_END = '%(message)s'
    _LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(
        self,
        options: ty.Optional[argparse.Namespace] = None,
        config: ty.Optional[cloud_config.CloudConfig] = None,
        **kwargs: ty.Any,
    ) -> None:
        context = {}
        if options:
            context = {
                'cloud': getattr(options, 'cloud', ''),
                'project': getattr(options, 'os_project_name', ''),
                'username': getattr(options, 'username', ''),
            }
        elif config:
            context = {
                'cloud': config.config.get('cloud', ''),
                'project': config.auth.get('project_name', ''),
                'username': config.auth.get('username', ''),
            }
        if context:
            self.fmt = (
                self._LOG_MESSAGE_BEGIN
                + (self._LOG_MESSAGE_CONTEXT % context)
                + self._LOG_MESSAGE_END
            )
        else:
            self.fmt = self._LOG_MESSAGE_BEGIN + self._LOG_MESSAGE_END
        logging.Formatter.__init__(self, self.fmt, self._LOG_DATE_FORMAT)


class LogConfigurator:
    _CONSOLE_MESSAGE_FORMAT = '%(message)s'

    def __init__(self, options: argparse.Namespace) -> None:
        self.root_logger = logging.getLogger('')
        self.root_logger.setLevel(logging.DEBUG)

        # Force verbose_level 3 on --debug
        self.dump_trace = False
        if options.debug:
            options.verbose_level = 3
            self.dump_trace = True

        # Always send higher-level messages to the console via stderr
        self.console_logger = logging.StreamHandler(sys.stderr)
        log_level = log_level_from_options(options)
        self.console_logger.setLevel(log_level)
        formatter = logging.Formatter(self._CONSOLE_MESSAGE_FORMAT)
        self.console_logger.setFormatter(formatter)
        self.root_logger.addHandler(self.console_logger)

        # Set the warning filter now
        set_warning_filter(log_level)

        # Set up logging to a file
        self.file_logger = None
        log_file = options.log_file
        if log_file:
            self.file_logger = logging.FileHandler(filename=log_file)
            self.file_logger.setFormatter(_FileFormatter(options=options))
            self.file_logger.setLevel(log_level)
            self.root_logger.addHandler(self.file_logger)

        # Requests logs some stuff at INFO that we don't want
        # unless we have DEBUG
        requests_log = logging.getLogger("requests")

        # Other modules we don't want DEBUG output for
        cliff_log = logging.getLogger('cliff')
        stevedore_log = logging.getLogger('stevedore')
        iso8601_log = logging.getLogger("iso8601")

        if options.debug:
            # --debug forces traceback
            requests_log.setLevel(logging.DEBUG)
        else:
            requests_log.setLevel(logging.ERROR)

        cliff_log.setLevel(logging.ERROR)
        stevedore_log.setLevel(logging.ERROR)
        iso8601_log.setLevel(logging.ERROR)

    def configure(self, cloud_config: cloud_config.CloudConfig) -> None:
        log_level = log_level_from_config(cloud_config.config)
        set_warning_filter(log_level)
        self.dump_trace = cloud_config.config.get('debug', self.dump_trace)
        self.console_logger.setLevel(log_level)

        log_file = cloud_config.config.get('log_file')
        if log_file:
            if not self.file_logger:
                self.file_logger = logging.FileHandler(filename=log_file)
            self.file_logger.setFormatter(_FileFormatter(config=cloud_config))
            self.file_logger.setLevel(log_level)
            self.root_logger.addHandler(self.file_logger)

        logconfig = cloud_config.config.get('logging')
        if logconfig:
            highest_level = logging.NOTSET
            for k in logconfig.keys():
                level = log_level_from_string(logconfig[k])
                logging.getLogger(k).setLevel(level)
                if highest_level < level:
                    highest_level = level
            self.console_logger.setLevel(highest_level)
            if self.file_logger:
                self.file_logger.setLevel(highest_level)
            # loggers that are not set will use the handler level, so we
            # need to set the global level for all the loggers
            for logkey in logging.Logger.manager.loggerDict.keys():
                logger = logging.getLogger(logkey)
                if logger.level == logging.NOTSET:
                    logger.setLevel(log_level)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Common client utilities"""

import argparse
import copy
import functools
import getpass
import logging
import os
import time
import typing as ty
import warnings

from cliff import columns as cliff_columns
from openstack import resource
from oslo_utils import importutils

from osc_lib import exceptions
from osc_lib.i18n import _

LOG = logging.getLogger(__name__)

_T = ty.TypeVar('_T')


def backward_compat_col_lister(
    column_headers: ty.List[str],
    columns: ty.List[str],
    column_map: ty.Dict[str, str],
) -> ty.List[str]:
    """Convert the column headers to keep column backward compatibility.

    Replace the new column name of column headers by old name, so that
    the column headers can continue to support to show the old column name by
    --column/-c option with old name, like: volume list -c 'Display Name'

    :param column_headers: The column headers to be output in list command.
    :param columns: The columns to be output.
    :param column_map: The key of map is old column name, the value is new
            column name, like: {'old_col': 'new_col'}
    """
    if not columns:
        return column_headers
    # NOTE(RuiChen): column_headers may be a tuple in some code, like:
    #                volume v1, convert it to a list in order to change
    #                the column name.
    column_headers = list(column_headers)
    for old_col, new_col in column_map.items():
        if old_col in columns:
            LOG.warning(
                _(
                    'The column "%(old_column)s" was deprecated, '
                    'please use "%(new_column)s" replace.'
                )
                % {'old_column': old_col, 'new_column': new_col}
            )
            if new_col in column_headers:
                column_headers[column_headers.index(new_col)] = old_col
    return column_headers


def backward_compat_col_showone(
    show_object: ty.MutableMapping[str, _T],
    columns: ty.List[str],
    column_map: ty.Dict[str, str],
) -> ty.MutableMapping[str, _T]:
    """Convert the output object to keep column backward compatibility.

    Replace the new column name of output object by old name, so that
    the object can continue to support to show the old column name by
    --column/-c option with old name, like: volume show -c 'display_name'

    :param show_object: The object to be output in create/show commands.
    :param columns: The columns to be output.
    :param column_map: The key of map is old column name, the value is new
        column name, like: {'old_col': 'new_col'}
    """
    if not columns:
        return show_object

    show_object = copy.deepcopy(show_object)
    for old_col, new_col in column_map.items():
        if old_col in columns:
            LOG.warning(
                _(
                    'The column "%(old_column)s" was deprecated, '
                    'please use "%(new_column)s" replace.'
                )
                % {'old_column': old_col, 'new_column': new_col}
            )
            if new_col in show_object:
                show_object.update({old_col: show_object.pop(new_col)})
    return show_object


def build_kwargs_dict(arg_name: str, value: _T) -> ty.Dict[str, _T]:
    """Return a dictionary containing `arg_name` if `value` is set."""
    kwargs = {}
    if value:
        kwargs[arg_name] = value
    return kwargs


def calculate_header_and_attrs(
    column_headers: ty.Sequence[str],
    attrs: ty.Sequence[str],
    parsed_args: argparse.Namespace,
) -> ty.Tuple[ty.Sequence[str], ty.Sequence[str]]:
    """Calculate headers and attribute names based on parsed_args.column.

    When --column (-c) option is specified, this function calculates
    column headers and expected API attribute names according to
    the OSC header/column definitions.

    This function also adjusts the content of parsed_args.columns
    if API attribute names are used in parsed_args.columns.
    This allows users to specify API attribute names in -c option.

    :param column_headers: A tuple/list of column headers to display
    :param attrs: a tuple/list of API attribute names. The order of
        corresponding column header and API attribute name must match.
    :param parsed_args: Parsed argument object returned by argparse parse_args
    :returns: A tuple of calculated headers and API attribute names.
    """
    if parsed_args.columns:
        header_attr_map = dict(zip(column_headers, attrs))
        expected_attrs = [
            header_attr_map.get(c, c) for c in parsed_args.columns
        ]
        attr_header_map = dict(zip(attrs, column_headers))
        expected_headers = [
            attr_header_map.get(c, c) for c in parsed_args.columns
        ]
        # If attribute name is used in parsed_args.columns
        # convert it into display names because cliff expects
        # name in parsed_args.columns and name in column_headers matches.
        parsed_args.columns = expected_headers
        return expected_headers, expected_attrs
    else:
        return column_headers, attrs


def env(*vars: str, **kwargs: ty.Any) -> ty.Optional[str]:
    """Search for the first defined of possibly many env vars

    Returns the first environment variable defined in vars, or
    returns the default defined in kwargs.
    """
    for v in vars:
        value = os.environ.get(v, None)
        if value:
            return value

    if 'default' in kwargs and kwargs['default'] is not None:
        return str(kwargs['default'])

    return None


def find_min_match(
    items: ty.Sequence[_T],
    sort_attr: str,
    **kwargs: ty.Any,
) -> ty.Sequence[_T]:
    """Find all resources meeting the given minimum constraints

    :param items: A List of objects to consider
    :param sort_attr: Attribute to sort the resulting list
    :param kwargs: A dict of attributes and their minimum values
    :rtype: A list of resources osrted by sort_attr that meet the minimums
    """

    def minimum_pieces_of_flair(item: _T) -> bool:
        """Find lowest value greater than the minumum"""

        result = True
        for k in kwargs:
            # AND together all of the given attribute results
            result = result and kwargs[k] <= get_field(item, k)
        return result

    return sort_items(list(filter(minimum_pieces_of_flair, items)), sort_attr)


# TODO(stephenfin): We should return a proper type, but how to do so without
# using generics? We should also deprecate this but there are a lot of users
# still.
def find_resource(
    manager: ty.Any,
    name_or_id: str,
    **kwargs: ty.Any,
) -> ty.Any:
    """Helper for the _find_* methods.

    :param manager: A client manager class
    :param name_or_id: The resource we are trying to find
    :param kwargs: To be used in calling .find()
    :rtype: The found resource

    This method will attempt to find a resource in a variety of ways.
    Primarily .get() methods will be called with `name_or_id` as an integer
    value, and tried again as a string value.

    If both fail, then a .find() is attempted, which is essentially calling
    a .list() function with a 'name' query parameter that is set to
    `name_or_id`.

    Lastly, if any kwargs are passed in, they will be treated as additional
    query parameters. This is particularly handy in the case of finding
    resources in a domain.

    """

    # Case 1: name_or_id is an ID, we need to call get() directly
    # for example: /projects/454ad1c743e24edcad846d1118837cac
    # For some projects, the name only will work. For keystone, this is not
    # enough information, and domain information is necessary.
    try:
        return manager.get(name_or_id)
    except Exception:  # noqa: S110
        pass

    if kwargs:
        # Case 2: name_or_id is a name, but we have query args in kwargs
        # for example: /projects/demo&domain_id=30524568d64447fbb3fa8b7891c10dd
        try:
            return manager.get(name_or_id, **kwargs)
        except Exception:  # noqa: S110
            pass

    # Case 3: Try to get entity as integer id. Keystone does not have integer
    # IDs, they are UUIDs, but some things in nova do, like flavors.
    try:
        if isinstance(name_or_id, int) or name_or_id.isdigit():
            return manager.get(int(name_or_id), **kwargs)
    # FIXME(dtroyer): The exception to catch here is dependent on which
    #                 client library the manager passed in belongs to.
    #                 Eventually this should be pulled from a common set
    #                 of client exceptions.
    except Exception as ex:
        if (
            type(ex).__name__ == 'NotFound'
            or type(ex).__name__ == 'HTTPNotFound'
            or type(ex).__name__ == 'TypeError'
        ):
            pass
        else:
            raise

    # Case 4: Try to use find.
    # Reset the kwargs here for find
    if len(kwargs) == 0:
        kwargs = {}

    try:
        # Prepare the kwargs for calling find
        if 'NAME_ATTR' in manager.resource_class.__dict__:
            # novaclient does this for oddball resources
            kwargs[manager.resource_class.NAME_ATTR] = name_or_id
        else:
            kwargs['name'] = name_or_id
    except Exception:  # noqa: S110
        pass

    # finally try to find entity by name
    try:
        return manager.find(**kwargs)
    # FIXME(dtroyer): The exception to catch here is dependent on which
    #                 client library the manager passed in belongs to.
    #                 Eventually this should be pulled from a common set
    #                 of client exceptions.
    except Exception as ex:
        if type(ex).__name__ == 'NotFound':
            msg = _("No %(resource)s with a name or ID of '%(id)s' exists.")
            raise exceptions.CommandError(
                msg
                % {
                    'resource': manager.resource_class.__name__.lower(),
                    'id': name_or_id,
                }
            )
        if type(ex).__name__ == 'NoUniqueMatch':
            msg = _(
                "More than one %(resource)s exists with the name '%(id)s'."
            )
            raise exceptions.CommandError(
                msg
                % {
                    'resource': manager.resource_class.__name__.lower(),
                    'id': name_or_id,
                }
            )
        else:
            pass

    # Case 5: For client with no find function, list all resources and hope
    # to find a matching name or ID.
    count = 0
    for res in manager.list():
        if res.get('id') == name_or_id or res.get('name') == name_or_id:
            count += 1
            _res = res
    if count == 0:
        # we found no match, report back this error:
        msg = _("Could not find resource %s")
        raise exceptions.CommandError(msg % name_or_id)
    elif count == 1:
        return _res
    else:
        # we found multiple matches, report back this error
        msg = _("More than one resource exists with the name or ID '%s'.")
        raise exceptions.CommandError(msg % name_or_id)


def format_dict(
    data: ty.Dict[str, ty.Any], prefix: ty.Optional[str] = None
) -> str:
    """Return a formatted string of key value pairs

    :param data: a dict
    :param prefix: the current parent keys in a recursive call
    :rtype: a string formatted to key='value'
    """

    if data is None:
        return None

    output = ""
    for s in sorted(data):
        if prefix:
            key_str = ".".join([prefix, s])
        else:
            key_str = s
        if isinstance(data[s], dict):
            # NOTE(dtroyer): Only append the separator chars here, quoting
            #                is completely handled in the terminal case.
            output = output + format_dict(data[s], prefix=key_str) + ", "
        elif data[s] is not None:
            output = output + key_str + "='" + str(data[s]) + "', "
        else:
            output = output + key_str + "=, "
    return output[:-2]


def format_dict_of_list(
    data: ty.Optional[ty.Dict[str, ty.List[ty.Any]]], separator: str = '; '
) -> ty.Optional[str]:
    """Return a formatted string of key value pair

    :param data: a dict, key is string, value is a list of string, for example:
                 {'public': ['2001:db8::8', '172.24.4.6']}
    :param separator: the separator to use between key/value pair
                      (default: '; ')
    :return: a string formatted to {'key1'=['value1', 'value2']} with separated
             by separator
    """
    if data is None:
        return None

    output = []
    for key in sorted(data):
        value = data[key]
        if value is None:
            continue
        value_str = format_list(value)
        group = f"{key}={value_str}"
        output.append(group)

    return separator.join(output)


def format_list(
    data: ty.Optional[ty.List[ty.Any]], separator: str = ', '
) -> ty.Optional[str]:
    """Return a formatted strings

    :param data: a list of strings
    :param separator: the separator to use between strings (default: ', ')
    :rtype: a string formatted based on separator
    """
    if data is None:
        return None

    return separator.join(sorted(data))


def format_list_of_dicts(
    data: ty.Optional[ty.List[ty.Dict[str, ty.Any]]],
) -> ty.Optional[str]:
    """Return a formatted string of key value pairs for each dict

    :param data: a list of dicts
    :rtype: a string formatted to key='value' with dicts separated by new line
    """
    if data is None:
        return None

    return '\n'.join(format_dict(i) for i in data)


def format_size(size: ty.Union[int, float, None]) -> str:
    """Display size of a resource in a human readable format

    :param size:
        The size of the resource in bytes.

    :returns:
        Returns the size in human-friendly format
    :rtype string:

    This function converts the size (provided in bytes) of a resource
    into a human-friendly format such as K, M, G, T, P, E, Z
    """

    suffix = ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']
    base = 1000.0
    index = 0

    size_ = float(size) if size is not None else 0.0
    while size_ >= base:
        index = index + 1
        size_ = size_ / base

    padded = f'{size_:.1f}'
    stripped = padded.rstrip('0').rstrip('.')

    return f'{stripped}{suffix[index]}'


def get_client_class(
    api_name: str,
    version: ty.Union[str, int, float],
    version_map: ty.Dict[str, ty.Type[_T]],
) -> ty.Any:
    """Returns the client class for the requested API version

    :param api_name: the name of the API, e.g. 'compute', 'image', etc
    :param version: the requested API version
    :param version_map: a dict of client classes keyed by version
    :rtype: a client class for the requested API version
    """
    warnings.warn(
        "This function is deprecated and is not necessary with openstacksdk."
        "Consider vendoring this if necessary.",
        category=DeprecationWarning,
    )

    try:
        client_path = version_map[str(version)]
    except (KeyError, ValueError):
        sorted_versions = sorted(
            version_map.keys(), key=lambda s: list(map(int, s.split('.')))
        )
        msg = _(
            "Invalid %(api_name)s client version '%(version)s'. "
            "must be one of: %(version_map)s"
        )
        raise exceptions.UnsupportedVersion(
            msg
            % {
                'api_name': api_name,
                'version': version,
                'version_map': ', '.join(sorted_versions),
            }
        )

    return importutils.import_class(client_path)


def get_dict_properties(
    item: ty.Dict[str, _T],
    fields: ty.Sequence[str],
    mixed_case_fields: ty.Optional[ty.Sequence[str]] = None,
    formatters: ty.Optional[  # type: ignore
        ty.Dict[str, ty.Type[cliff_columns.FormattableColumn]]
    ] = None,
) -> ty.Tuple[ty.Any, ...]:
    """Return a tuple containing the item properties.

    :param item: a single dict resource
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
       to format the values
    """
    if mixed_case_fields is None:
        mixed_case_fields = []
    if formatters is None:
        formatters = {}

    row = []

    for field in fields:
        if field in mixed_case_fields:
            field_name = field.replace(' ', '_')
        else:
            field_name = field.lower().replace(' ', '_')
        data: ty.Any = item[field_name] if field_name in item else ''
        if field in formatters:
            formatter = formatters[field]
            # columns must be either a subclass of FormattableColumn
            if isinstance(formatter, type) and issubclass(
                formatter, cliff_columns.FormattableColumn
            ):
                data = formatter(data)
            # or a partial wrapping one (to allow us to pass extra parameters)
            elif (
                isinstance(formatter, functools.partial)
                and isinstance(formatter.func, type)
                and issubclass(formatter.func, cliff_columns.FormattableColumn)
            ):
                data = formatter(data)
            # otherwise it's invalid
            else:
                msg = "Invalid formatter provided."
                raise exceptions.CommandError(msg)

        row.append(data)
    return tuple(row)


def get_item_properties(
    item: ty.Dict[str, _T],
    fields: ty.Sequence[str],
    mixed_case_fields: ty.Optional[ty.Sequence[str]] = None,
    formatters: ty.Optional[  # type: ignore
        ty.Dict[str, ty.Type[cliff_columns.FormattableColumn]]
    ] = None,
) -> ty.Tuple[ty.Any, ...]:
    """Return a tuple containing the item properties.

    :param item: a single item resource (e.g. Server, Project, etc)
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
       to format the values
    """
    if mixed_case_fields is None:
        mixed_case_fields = []
    if formatters is None:
        formatters = {}

    row = []

    for field in fields:
        if field in mixed_case_fields:
            field_name = field.replace(' ', '_')
        else:
            field_name = field.lower().replace(' ', '_')
        data = getattr(item, field_name, '')
        if field in formatters:
            formatter = formatters[field]
            # columns must be either a subclass of FormattableColumn
            if isinstance(formatter, type) and issubclass(
                formatter, cliff_columns.FormattableColumn
            ):
                data = formatter(data)
            # or a partial wrapping one (to allow us to pass extra parameters)
            elif (
                isinstance(formatter, functools.partial)
                and isinstance(formatter.func, type)
                and issubclass(formatter.func, cliff_columns.FormattableColumn)
            ):
                data = formatter(data)
            # otherwise it's invalid
            else:
                msg = "Invalid formatter provided."
                raise exceptions.CommandError(msg)

        row.append(data)
    return tuple(row)


def get_effective_log_level() -> int:
    """Returns the lowest logging level considered by logging handlers

    Retrieve and return the smallest log level set among the root
    logger's handlers (in case of multiple handlers).
    """
    root_log = logging.getLogger()
    min_log_lvl = logging.CRITICAL
    for handler in root_log.handlers:
        min_log_lvl = min(min_log_lvl, handler.level)
    return min_log_lvl


def get_field(item: _T, field: str) -> ty.Any:
    try:
        if isinstance(item, dict):
            return item[field]
        else:
            return getattr(item, field)
    except Exception:
        msg = _("Resource doesn't have field %s")
        raise exceptions.CommandError(msg % field)


def get_password(
    stdin: ty.TextIO,
    prompt: ty.Optional[str] = None,
    confirm: bool = True,
) -> str:
    message = prompt or "User Password:"
    if hasattr(stdin, 'isatty') and stdin.isatty():
        try:
            while True:
                first_pass = getpass.getpass(message)
                if not confirm:
                    return first_pass
                second_pass = getpass.getpass("Repeat " + message)
                if first_pass == second_pass:
                    return first_pass
                msg = _("The passwords entered were not the same")
                print(msg)
        except EOFError:  # Ctl-D
            msg = _("Error reading password")
            raise exceptions.CommandError(msg)
    msg = _("No terminal detected attempting to read password")
    raise exceptions.CommandError(msg)


def is_ascii(string: ty.Union[str, bytes]) -> bool:
    try:
        if isinstance(string, bytes):
            string.decode('ascii')
        else:
            string.encode('ascii')
        return True
    except (UnicodeEncodeError, UnicodeDecodeError):
        return False


def read_blob_file_contents(blob_file: str) -> str:
    try:
        with open(blob_file) as file:
            blob = file.read().strip()
        return blob
    except OSError:
        msg = _("Error occurred trying to read from file %s")
        raise exceptions.CommandError(msg % blob_file)


def sort_items(
    items: ty.Sequence[_T],
    sort_str: str,
    sort_type: ty.Optional[ty.Type[ty.Any]] = None,
) -> ty.Sequence[_T]:
    """Sort items based on sort keys and sort directions given by sort_str.

    :param items: a list or generator object of items
    :param sort_str: a string defining the sort rules, the format is
        '<key1>:[direction1],<key2>:[direction2]...', direction can be 'asc'
        for ascending or 'desc' for descending, if direction is not given,
        it's ascending by default
    :return: sorted items
    """
    if not sort_str:
        return items
    # items may be a generator object, transform it to a list
    items = list(items)
    sort_keys = sort_str.strip().split(',')
    for sort_key in reversed(sort_keys):
        reverse = False
        if ':' in sort_key:
            sort_key, direction = sort_key.split(':', 1)
            if not sort_key:
                msg = _("'<empty string>'' is not a valid sort key")
                raise exceptions.CommandError(msg)
            if direction not in ['asc', 'desc']:
                if not direction:
                    direction = "<empty string>"
                msg = _(
                    "'%(direction)s' is not a valid sort direction for "
                    "sort key %(sort_key)s, use 'asc' or 'desc' instead"
                )
                raise exceptions.CommandError(
                    msg
                    % {
                        'direction': direction,
                        'sort_key': sort_key,
                    }
                )
            if direction == 'desc':
                reverse = True

        def f(x: ty.Any) -> ty.Any:
            # Attempts to convert items to same 'sort_type' if provided.
            # This is due to Python 3 throwing TypeError if you attempt to
            # compare different types
            item = get_field(x, sort_key)
            if sort_type and not isinstance(item, sort_type):
                try:
                    item = sort_type(item)
                except Exception:
                    # Can't convert, so no sensible way to compare
                    item = sort_type()
            return item

        items.sort(key=f, reverse=reverse)

    return items


def wait_for_delete(
    manager: ty.Any,
    res_id: str,
    status_field: str = 'status',
    error_status: ty.Sequence[str] = ['error'],
    exception_name: ty.Sequence[str] = ['NotFound'],
    sleep_time: int = 5,
    timeout: int = 300,
    callback: ty.Optional[ty.Callable[[int], None]] = None,
) -> bool:
    """Wait for resource deletion

    :param manager: the manager from which we can get the resource
    :param res_id: the resource id to watch
    :param status_field: the status attribute in the returned resource object,
        this is used to check for error states while the resource is being
        deleted
    :param error_status: a list of status strings for error
    :param exception_name: a list of exception strings for deleted case
    :param sleep_time: wait this long between checks (seconds)
    :param timeout: check until this long (seconds)
    :param callback: called per sleep cycle, useful to display progress; this
        function is passed a progress value during each iteration of the wait
        loop
    :rtype: True on success, False if the resource has gone to error state or
        the timeout has been reached
    """
    warnings.warn(
        "This function is deprecated as it does not support openstacksdk "
        "clients. Consider vendoring this if necessary.",
        category=DeprecationWarning,
    )

    total_time = 0
    while total_time < timeout:
        try:
            # might not be a bad idea to re-use find_resource here if it was
            # a bit more friendly in the exceptions it raised so we could just
            # handle a NotFound exception here without parsing the message
            res = manager.get(res_id)
        except Exception as ex:
            if type(ex).__name__ in exception_name:
                return True
            raise

        status = getattr(res, status_field, '').lower()
        if status in error_status:
            return False

        if callback:
            progress = getattr(res, 'progress', None) or 0
            callback(progress)
        time.sleep(sleep_time)
        total_time += sleep_time

    # if we got this far we've timed out
    return False


def wait_for_status(
    status_f: ty.Callable[[str], object],
    res_id: str,
    status_field: str = 'status',
    success_status: ty.Sequence[str] = ['active'],
    error_status: ty.Sequence[str] = ['error'],
    sleep_time: int = 5,
    callback: ty.Optional[ty.Callable[[int], None]] = None,
) -> bool:
    """Wait for status change on a resource during a long-running operation

    :param status_f: a status function that takes a single id argument
    :param res_id: the resource id to watch
    :param status_field: the status attribute in the returned resource object
    :param success_status: a list of status strings for successful completion
    :param error_status: a list of status strings for error
    :param sleep_time: wait this long (seconds)
    :param callback: called per sleep cycle, useful to display progress
    :rtype: True on success
    """
    warnings.warn(
        "This function is deprecated as it does not support openstacksdk "
        "clients. Consider vendoring this if necessary.",
        category=DeprecationWarning,
    )

    while True:
        res = status_f(res_id)
        status = getattr(res, status_field, '').lower()
        if status in success_status:
            retval = True
            break
        elif status in error_status:
            retval = False
            break
        if callback:
            progress = getattr(res, 'progress', None) or 0
            callback(progress)
        time.sleep(sleep_time)
    return retval


def get_osc_show_columns_for_sdk_resource(
    sdk_resource: resource.Resource,
    osc_column_map: ty.Dict[str, str],
    invisible_columns: ty.Optional[ty.Sequence[str]] = None,
) -> ty.Tuple[ty.Tuple[str, ...], ty.Tuple[str, ...]]:
    """Get and filter the display and attribute columns for an SDK resource.

    Common utility function for preparing the output of an OSC show command.
    Some of the columns may need to get renamed, others made invisible.

    :param sdk_resource: An SDK resource
    :param osc_column_map: A hash of mappings for display column names
    :param invisible_columns: A list of invisible column names

    :returns: Two tuples containing the names of the display and attribute
              columns
    """

    if getattr(sdk_resource, 'allow_fetch', None) is not None:
        # OSC at the moment lands here with FakeResource objects which are not
        # 100% sdk compatible. Unless we introduce SDK test/fake resources we
        # should check presence of the specific method
        resource_dict = sdk_resource.to_dict(
            body=True, headers=False, ignore_none=False
        )
    else:
        # We might land here with not a real SDK Resource (during the
        # transition period).
        resource_dict = sdk_resource

    # Build the OSC column names to display for the SDK resource.
    attr_map = {}
    display_columns = list(resource_dict.keys())
    invisible_columns = [] if invisible_columns is None else invisible_columns
    for col_name in invisible_columns:
        if col_name in display_columns:
            display_columns.remove(col_name)
    for sdk_attr, osc_attr in osc_column_map.items():
        if sdk_attr in display_columns:
            attr_map[osc_attr] = sdk_attr
            display_columns.remove(sdk_attr)
        if osc_attr not in display_columns:
            display_columns.append(osc_attr)
    sorted_display_columns = sorted(display_columns)

    # Build the SDK attribute names for the OSC column names.
    attr_columns = []
    for column in sorted_display_columns:
        new_column = attr_map[column] if column in attr_map else column
        attr_columns.append(new_column)
    return tuple(sorted_display_columns), tuple(attr_columns)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Test Timing pseudo-command"""

import datetime

from keystoneauth1 import session

from osc_lib.command import timing
from osc_lib.tests import fakes
from osc_lib.tests import utils


timing_method = 'GET'
timing_url = 'http://localhost:5000'
timing_elapsed = 0.872809


class FakeGenericClient:
    def __init__(self, **kwargs):
        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']


class TestTiming(utils.TestCommand):
    columns = (
        'URL',
        'Seconds',
    )

    def setUp(self):
        super().setUp()

        self.app.timing_data = []

        self.app.client_manager.compute = FakeGenericClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )

        self.app.client_manager.volume = FakeGenericClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )

        # Get the command object to test
        self.cmd = timing.Timing(self.app, None)

    def test_timing_list_no_data(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(self.columns, columns)
        datalist = [
            (
                'Total',
                0.0,
            )
        ]
        self.assertEqual(datalist, data)

    def test_timing_list(self):
        self.app.timing_data = [
            session.RequestTiming(
                method=timing_method,
                url=timing_url,
                elapsed=datetime.timedelta(
                    microseconds=timing_elapsed * 1000000
                ),
            )
        ]

        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(self.columns, columns)
        datalist = [
            (timing_method + ' ' + timing_url, timing_elapsed),
            ('Total', timing_elapsed),
        ]
        self.assertEqual(datalist, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse
import typing as ty

from openstack import connection
from openstack import exceptions
from openstack.identity.v3 import project

from osc_lib.i18n import _


def add_project_owner_option_to_parser(
    parser: argparse.ArgumentParser,
) -> None:
    """Register project and project domain options.

    :param parser: argparse.Argument parser object.
    """
    parser.add_argument(
        '--project',
        metavar='<project>',
        help=_("Owner's project (name or ID)"),
    )
    parser.add_argument(
        '--project-domain',
        metavar='<project-domain>',
        help=_(
            'Domain the project belongs to (name or ID). '
            'This can be used in case collisions between project names '
            'exist.'
        ),
    )


# TODO(stephenfin): This really doesn't belong here. This should be part of
# openstacksdk itself.
def find_project(
    sdk_connection: connection.Connection,
    name_or_id: str,
    domain_name_or_id: ty.Optional[str] = None,
) -> project.Project:
    """Find a project by its name name or ID.

    If Forbidden to find the resource (a common case if the user does not have
    permission), then return the resource by creating a local instance of
    openstack.identity.v3.Project resource.

    :param sdk_connection: Connection object of OpenStack SDK.
    :type sdk_connection: `openstack.connection.Connection`
    :param name_or_id: Name or ID of the project
    :type name_or_id: string
    :param domain_name_or_id: Domain name or ID of the project.
        This can be used when there are multiple projects with a same name.
    :returns: the project object found
    :rtype: `openstack.identity.v3.project.Project`
    """
    try:
        if domain_name_or_id:
            domain = sdk_connection.identity.find_domain(
                domain_name_or_id, ignore_missing=False
            )
            domain_id = domain.id
        else:
            domain_id = None
        return sdk_connection.identity.find_project(
            name_or_id, ignore_missing=False, domain_id=domain_id
        )
    # NOTE: OpenStack SDK raises HttpException for 403 response code.
    # There is no specific exception class at now, so we need to catch
    # HttpException and check the status code.
    except exceptions.HttpException as e:
        if e.status_code == 403:
            return project.Project(id=name_or_id, name=name_or_id)
        raise
